@@ -42,6 +42,8 @@
 #include <string.h>
 #include <candl/candl.h>
 
+#include <candl/piplib-wrapper.h>
+
 #include <assert.h>
 
 #ifdef CANDL_SUPPORTS_ISL
@@ -52,6 +54,8 @@
 # include <isl/set.h>
 #endif
 
+
+
 /******************************************************************************
  *                          Structure display function                        *
  ******************************************************************************/
@@ -1152,8 +1156,6 @@ candl_dependence_p candl_dependence_system(CandlStatement* source,
 {
   candl_dependence_p dependence = NULL;
   CandlMatrix * system;
-  PipOptions * options;
-  PipQuast * solution;
 
   /* First, a trivial case: for two different statements at depth 0, there is
    * a dependence only if the source is textually before the target.
@@ -1172,14 +1174,7 @@ candl_dependence_p candl_dependence_system(CandlStatement* source,
   if (! candl_dependence_gcd_test(source, target, system, depth))
     return NULL;
 
-  options = pip_options_init();
-  options->Simplify = 1;
-  options->Urs_parms = -1;
-  options->Urs_unknowns = -1;
-  solution = pip_solve(system,context, -1, options);
-
-  if ((solution != NULL) &&
-      ((solution->list != NULL) || (solution->condition != NULL)))
+  if (pip_has_rational_point (system, context, 1))
     {
       dependence = candl_dependence_malloc();
 
@@ -1192,11 +1187,8 @@ candl_dependence_p candl_dependence_system(CandlStatement* source,
       dependence->ref_target = ref_t;
       dependence->domain     = system;
     }
-  else {
+  else 
     candl_matrix_free(system);
-  }
-  pip_options_free(options);
-  pip_quast_free(solution);
 
   return dependence;
 }
@@ -1374,6 +1366,10 @@ candl_dependence_p candl_dependence(candl_program_p program,
       candl_compute_last_writer(dependence, program);
   }
 
+  /* Remove some transitively covered dependences (experimental). */
+  if (options->prune_dups)
+    dependence = candl_dependence_prune_transitively_covered (dependence);
+
   return dependence;
 }
 
@@ -1589,8 +1585,8 @@ candl_dependence_compute_lb (CandlMatrix* m, Entier* lb, int col)
       ((solution->list != NULL) || (solution->condition != NULL)))
     {
       l = solution->list;
-      for (; l && col > 0; l = l->next, --col)
-	;
+      while (col-- > 1)
+	l = l->next;
       CANDL_assign(*lb, l->vector->the_vector[0]);
     }
   pip_options_free (options);
@@ -1615,7 +1611,6 @@ candl_dependence_check_domain_is_included(CandlStatement* s1,
   Entier lb; CANDL_init(lb);
   max = s1->depth < max ? s1->depth : max;
   max = s2->depth < max ? s2->depth : max;
-
   CandlMatrix* m = candl_matrix_malloc(s2->domain->NbRows + s2->depth - max +1,
 				       s2->domain->NbColumns);
   int i, j;
@@ -1999,8 +1994,6 @@ candl_dependence_is_loop_carried (candl_program_p program,
   if (j != i)
     return 0;
 
-  /* Ensure it is not a basic loop-independent dependence (pure equality of
-     the access functions, and contain the loop iterator. */
   int k = 0;
   int l = 0;
   CandlMatrix* mats;
@@ -2013,50 +2006,159 @@ candl_dependence_is_loop_carried (candl_program_p program,
     matt = dep->target->read;
   else
     matt = dep->target->written;
-  int has_iter = 0;
+
+  int loopdep = 1;
+  int src_ref_iter = 0;
+  int dst_ref_iter = 0;
+  for (k = 0; k == 0 ||
+	 (dep->ref_source + k < mats->NbRows &&
+	  CANDL_get_si(mats->p[dep->ref_source + k][0]) == 0); ++k)
+    if (CANDL_get_si(mats->p[dep->ref_source + k][i + 1]) != 0)
+      {
+	loopdep = 0;
+	src_ref_iter = 1;
+	break;
+      }
+  for (k = 0; k == 0 ||
+	 (dep->ref_target + k < matt->NbRows &&
+	  CANDL_get_si(matt->p[dep->ref_target + k][0]) == 0); ++k)
+    if (CANDL_get_si(matt->p[dep->ref_target + k][j + 1]) != 0)
+      {
+	loopdep = 0;
+	dst_ref_iter = 1;
+	break;
+      }
+
+  /* Ensure it is not a basic loop-independent dependence (pure
+     equality of the access functions for the surrounding iterators +
+     parameters + constant, no occurence of the inner loop iterators,
+     and contain the current loop iterator. */
+  int must_test = 0;
+  k = 0;
   do
     {
-      // Ensure the access functions are equal (conservative).
-      for (l = 0; l < mats->NbColumns; ++l)
-	if (CANDL_get_si(mats->p[dep->ref_source + k][l]) !=
-	    CANDL_get_si(matt->p[dep->ref_target + k][l]))
-	  break;
       // Ensure the reference do reference the current loop iterator
       // to be tested.
       if (CANDL_get_si(mats->p[dep->ref_source + k][i + 1]) != 0)
-	has_iter = 1;
+	{
+	  int kp = 0;
+	  do
+	    {
+	      if (CANDL_get_si(matt->p[dep->ref_target + kp][j + 1]) != 0)
+		{
+		  // Ensure the access functions are equal for the
+		  // surrounding loop iterators, and no inner iterator
+		  // is referenced.
+		  for (l = 0; l <= i + 1; ++l)
+		    if (CANDL_get_si(mats->p[dep->ref_source + k][l]) !=
+		  	CANDL_get_si(matt->p[dep->ref_target + kp][l]))
+		      {
+		  	must_test = 1;
+		  	break;
+		      }
+		  int m = l;
+		  if (!must_test)
+		    for (; l <= dep->source->depth; ++l)
+		      if (CANDL_get_si(mats->p[dep->ref_source + k][l]) != 0)
+		  	{
+		  	  must_test = 1;
+		  	  break;
+		  	}
+		  if (!must_test)
+		    for (; m <= dep->target->depth; ++m)
+		      if (CANDL_get_si(matt->p[dep->ref_target + kp][m]) != 0)
+		  	{
+		  	  must_test = 1;
+		  	  break;
+		  	}
+		  if (!must_test)
+		    for (; l < mats->NbColumns; ++l, ++m)
+		      if (CANDL_get_si(mats->p[dep->ref_source + k][l]) !=
+		  	  CANDL_get_si(matt->p[dep->ref_target + kp][m]))
+		  	{
+		  	  must_test = 1;
+		  	  break;
+		  	}
+		}
+	      ++kp;
+	    }
+	  while (!must_test &&
+		 dep->ref_target + kp < matt->NbRows &&
+		 CANDL_get_si(matt->p[dep->ref_target + kp][0]) == 0);
+	}
       ++k;
     }
-  while (dep->ref_source + k < mats->NbRows &&
+  while (!must_test &&
+	 dep->ref_source + k < mats->NbRows &&
 	 CANDL_get_si(mats->p[dep->ref_source + k][0]) == 0);
-  if (l == mats->NbColumns && has_iter)
+  if (src_ref_iter && dst_ref_iter && !must_test)
     return 0;
 
-  /* Final check. The dependence exists only because the loop
-     iterates. Make the loop not iterate and check if there's still
-     dependent iterations. */
-  CandlMatrix* m = candl_matrix_malloc(dep->domain->NbRows + 2,
-				       dep->domain->NbColumns);
-  CANDL_set_si(m->p[m->NbRows - 2][i + 1], -1);
-  CANDL_set_si(m->p[m->NbRows - 1][dep->source->depth + 1 + j], -1);
-  /* Copy the rest of the matrix. */
-  int ii, jj;
-  for (ii = 0; ii < dep->domain->NbRows; ++ii)
-    for (jj = 0; jj < dep->domain->NbColumns; ++jj)
-      CANDL_assign(m->p[ii][jj], dep->domain->p[ii][jj]);
-  /* Compute real lb of loops. */
-  Entier lb; CANDL_init(lb);
-  candl_dependence_compute_lb (m, &lb, i + 1);
-  CANDL_assign(m->p[m->NbRows - 2][m->NbColumns - 1], lb);
-  candl_dependence_compute_lb (m, &lb, dep->source->depth + 1 + j);
-  CANDL_assign(m->p[m->NbRows - 1][m->NbColumns - 1], lb);
-  int ret = candl_matrix_check_point(m, program->context);
-  CANDL_clear(lb);
 
-  /* Be clean. */
-  candl_matrix_free(m);
+  /* Final check. For loop i, the dependence is loop carried if there exists
+     x_i^R != x_i^S in the dependence polyhedron, with
+     x_{1..i-1}^R = x_{1..i-1}^S
+  */
+  int pos;
+  CandlMatrix* mat = dep->domain;
+  CandlStatement* src = dep->source;
+  CandlMatrix* testsyst =
+    candl_matrix_malloc(mat->NbRows + 1 + dep->source->depth, mat->NbColumns);
+  for (pos = 0; pos < mat->NbRows; ++pos)
+    for (j = 0; j < mat->NbColumns; ++j)
+      CANDL_assign(testsyst->p[pos][j], mat->p[pos][j]);
+  for (j = 0; j < i; ++j)
+    {
+      CANDL_set_si(testsyst->p[pos + j + 1][0], 0);
+      CANDL_set_si(testsyst->p[pos + j + 1][1 + j], -1);
+      CANDL_set_si(testsyst->p[pos + j + 1][1 + j + dep->source->depth], 1);
+    }
 
-  return !ret;
+  int has_pt = 0;
+  // Test for '>'.
+  CANDL_set_si(testsyst->p[pos][0], 1);
+  CANDL_set_si(testsyst->p[pos][testsyst->NbColumns - 1], -1);
+  CANDL_set_si(testsyst->p[pos][1 + i], 1);
+  CANDL_set_si(testsyst->p[pos][1 + i + src->depth], -1);
+  has_pt = pip_has_rational_point (testsyst, NULL, 1);
+  if (! has_pt)
+    {
+      // Test for '<'.
+      CANDL_set_si(testsyst->p[pos][1 + i], -1);
+      CANDL_set_si(testsyst->p[pos][1 + i + src->depth], 1);
+      has_pt = pip_has_rational_point (testsyst, NULL, 1);
+    }
+
+  candl_matrix_free (testsyst);
+
+  return has_pt;
+
+  /* LNP: OLD VERSION. The above is more robust. */
+/*   /\* Final check. The dependence exists only because the loop */
+/*      iterates. Make the loop not iterate and check if there's still */
+/*      dependent iterations. *\/ */
+/*   CandlMatrix* m = candl_matrix_malloc(dep->domain->NbRows + 2, */
+/* 				       dep->domain->NbColumns); */
+/*   CANDL_set_si(m->p[m->NbRows - 2][i + 1], -1); */
+/*   CANDL_set_si(m->p[m->NbRows - 1][dep->source->depth + 1 + j], -1); */
+/*   /\* Copy the rest of the matrix. *\/ */
+/*   int ii, jj; */
+/*   for (ii = 0; ii < dep->domain->NbRows; ++ii) */
+/*     for (jj = 0; jj < dep->domain->NbColumns; ++jj) */
+/*       CANDL_assign(m->p[ii][jj], dep->domain->p[ii][jj]); */
+/*   /\* Compute real lb of loops. *\/ */
+/*   Entier lb; CANDL_init(lb); */
+/*   candl_dependence_compute_lb (m, &lb, i + 1); */
+/*   CANDL_assign(m->p[m->NbRows - 2][m->NbColumns - 1], lb); */
+/*   candl_dependence_compute_lb (m, &lb, dep->source->depth + 1 + j); */
+/*   CANDL_assign(m->p[m->NbRows - 1][m->NbColumns - 1], lb); */
+/*   int ret = candl_matrix_check_point(m, program->context); */
+/*   CANDL_clear(lb); */
+
+/*   /\* Be clean. *\/ */
+/*   candl_matrix_free(m); */
+
+/*   return !ret; */
 }
 
 
@@ -2240,7 +2342,6 @@ candl_dependence_analyze_scalars(candl_program_p program,
 	      /* No more statement in the chain, exit. */
 	      if (stmts[0] == NULL)
 		break;
-
 	      int c = 0;
 	      is_priv = candl_dependence_var_is_ref (stmts[c], scalars[i])
 		== CANDL_VAR_IS_DEF;
@@ -2310,16 +2411,16 @@ candl_dependence_analyze_scalars(candl_program_p program,
 			{
 			  free (program->scalars_privatizable);
 			  program->scalars_privatizable =
-			    (int*)malloc(priv_buff_size * 2 * sizeof(int));
+			    (int*)malloc(priv_buff_size * sizeof(int));
 			  for (l = 0; l < priv_buff_size; ++l)
 			    program->scalars_privatizable[l] = -1;
 			}
-		      if (2 * nb_priv == priv_buff_size)
+		      if (nb_priv == priv_buff_size)
 			{
 			  program->scalars_privatizable =
 			    realloc(program->scalars_privatizable,
 				    (priv_buff_size *= 2) * sizeof(int));
-			  for (l = 2 * nb_priv; l < priv_buff_size; ++l)
+			  for (l = nb_priv; l < priv_buff_size; ++l)
 			    program->scalars_privatizable[l] = -1;
 			}
 		      /* Memorize the scalar information in the
@@ -2605,7 +2706,6 @@ int candl_dep_compute_lastwriter (CandlDependence **dep, CandlProgram *prog)
     return 0;
 }
 
-
 /**
  * Compute the last writer for each RAW, WAW, and RAR dependence. This will
  * modify the dependence polyhedra. Be careful of any references to the old
